---------------------------------------------------------------------------------
-- FILELIB.E
-- version 0.3a
-- written by Jason Mirwald
-- LAST MODIFIED : 09/21/2002
---------------------------------------------------------------------------------

-- NOTE: max number of files that can be open at one time
-- by Euphoria is 25 including 0,1,2 (in/out/error)

----------------------------------------------

global procedure seekfn(integer fn, atom byte_position)
 if fn > 0 then
  machine_proc(19, {fn, byte_position})
 end if
end procedure

----------------------------------------------

global function wherefn(integer fn)
 return machine_func(20, fn)
end function

----------------------------------------------

-- modified version of getc()
-- takes args as; getchar({fn,number}) as well as getchar(fn)

global function getchar( object o )
 sequence s
 integer fn
 s = {}
 if atom(o) then
  return getc(o)
 end if
 if length(o) = 2 then
  fn = o[1]
  for n = 1 to o[2] do
   s &= getc(fn)
  end for
  return s
 end if
 return -1
end function

----------------------------------------------

global procedure put2( integer fn, object o )
 sequence s
 atom a
 s = {}
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  a = o[n]
  if a < 0 then
   a = #FFFF+a+1
  end if
  puts(fn,{remainder(a,#100),floor(a/#100)})
 end for
end procedure

----------------------------------------------

global procedure put4(integer fn, object o)
 atom a
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  a = o[n]
  if a < 0 then
   a = #FFFFFFFF+a+1
  end if
  puts(fn,and_bits(#FF,{a,a/#100,a/#10000,a/#1000000}))
 end for
end procedure

----------------------------------------------

global procedure put_float32(integer fn, object o)
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  puts(fn,machine_func(48,o[n]))
 end for
end procedure

----------------------------------------------

global procedure put_float64(integer fn, object o)
 if atom(o) then
  o = {o}
 end if
 for n = 1 to length(o) do
  puts(fn,machine_func(46,o[n]))
 end for
end procedure

----------------------------------------------

global function get2u(object o)
 sequence s
 integer fn
 atom a
 s = {}
 if atom(o) then
  a = getc(o)
  a += (getc(o)*#0100)
  return a
 else
  if length(o) = 2 then
   fn = o[1]
   for n = 1 to o[2] do
    a = getc(fn)
    a += (getc(o)*#0100)
    s &= a
   end for
   return s
  end if
 end if
 return -1
end function

----------------------------------------------

global function get2s(object o)
 sequence s
 integer fn
 atom a
 s = {}
 if atom(o) then
  a = getc(o)
  a += (getc(o)*#0100)
  return a-(#10000*(and_bits(a,#8000)!=0))
 else
  if length(o) = 2 then
   fn = o[1]
   for n = 1 to o[2] do
    a = getc(fn)
    a += (getc(fn)*#0100)
    s &= a-(#10000*(and_bits(a,#8000)!=0))
   end for
   return s
  end if
 end if
 return -1
end function

----------------------------------------------

global function get4u(object o)
 sequence s
 integer fn
 atom a
 s = {}
 if atom(o) then
  a = getc(o)
  a += (getc(o)*#0100)
  a += (getc(o)*#010000)
  a += (getc(o)*#01000000)
  return a
 else
  if length(o) = 2 then
   fn = o[1]
   for n = 1 to o[2] do
    a = getc(fn)
    a += (getc(fn)*#0100)
    a += (getc(fn)*#010000)
    a += (getc(fn)*#01000000)
    s &= a
   end for
   return s
  end if
 end if
 return -1
end function

----------------------------------------------

global function get4s(object o)
 sequence s
 integer fn
 atom a
 s = {}
 if atom(o) then
  a = getc(o)
  a += (getc(o)*#0100)
  a += (getc(o)*#010000)
  a += (getc(o)*#01000000)
  return a-(#100000000*(and_bits(a,#80000000)!=0))
 else
  if length(o) = 2 then
   fn = o[1]
   for n = 1 to o[2] do
    a = getc(fn)
    a += (getc(fn)*#0100)
    a += (getc(fn)*#010000)
    a += (getc(fn)*#01000000)
    s &= a-(#100000000*(and_bits(a,#80000000)!=0))
   end for
   return s
  end if
 end if
 return -1
end function

----------------------------------------------

global function get_float32(object o)
 sequence s
 integer fn
 s = {}
 if atom(o) then
  return machine_func(49, getchar({o,4}) )
 else
  if length(o) = 2 then
   fn = o[1]
   for n = 1 to o[2] do
    s &= machine_func(49, getchar({fn,4}) )
   end for
   return s
  end if
 end if
 return -1
end function

----------------------------------------------

global function get_float64(object o)
 sequence s
 integer fn
 s = {}
 if atom(o) then
  return machine_func(47, getchar({o,8}) )
 else
  if length(o) = 2 then
   fn = o[1]
   for n = 1 to o[2] do
    s &= machine_func(47, getchar({fn,8}) )
   end for
   return s
  end if
 end if
 return -1
end function

----------------------------------------------

-- this *always* returns a seqence

global function get_string(integer file)
 atom c
 sequence s
 s = {}
 c = getc(file)
 while c != -1 do
  s &= c
  if c = 10 then
   exit
  end if
  c = getc(file)
 end while
 return s
end function

----------------------------------------------

global function file_size( integer file )
 atom sz, loc
 sz = 0
 if file > 0 then
  loc = wherefn(file)
  seekfn(file,-1)
  sz = wherefn(file)
  seekfn(file,loc)
 end if
 return sz
end function

----------------------------------------------

-- storage and retrieval of binary objects in files
-- all float values are stored in IEEE 64-bit format
-- to preserve accuracy

constant
 uchar = 1,
 char = 2,
 word = 3,
 short = 4,
 dword = 5,
 long = 6,
 double = 7,
 string = 8,
 beg_seq = 9,
 end_seq = 10,
 beg_obj = 11,
 end_obj = 12

type char_string(object o)
 atom r
 r = 1
 if sequence(o) then
  for n = 1 to length(o) do
   if integer(o[n]) then
    if o[n] < 0 or o[n] > 255 then
     r = 0
     exit
    end if
   else
    r = 0
    exit
   end if
  end for
 end if
 return r
end type


function atom_size( atom a )
 if a-floor(a) = 0 then
  if a >= 0 then
   if a <= #FF then
    return uchar
   elsif a <= #FFFF then
    return word
   elsif a <= #FFFFFFFF then
    return dword
   else
    return double
   end if
  else -- negative
   if a >= 0-#7F then
    return char
   elsif a >= 0-#7FFF then
    return short
   elsif a >= 0-#7FFFFFFF then
    return long
   else
    return double
   end if
  end if
 else -- float
  return double
 end if
 return 0
end function


procedure put_object_l( integer fn, object o )
 atom sz
 if atom(o) then
  sz = atom_size(o)
  puts(fn,sz)
  if sz = uchar or sz = char then
   puts(fn,o)
  elsif sz = word or sz = short then
   put2(fn,o)
  elsif sz = dword or sz = long then
   put4(fn,o)
  elsif sz = double then
   put_float64(fn,o)
  end if
 elsif char_string(o) then
  puts(fn,string)
  put4(fn,length(o))
  puts(fn,o)
 else
  puts(fn,beg_seq)
  for n = 1 to length(o) do
   put_object_l(fn,o[n])
  end for
  puts(fn,end_seq)
 end if
end procedure


global procedure put_object( integer fn, object o )
 puts(fn,beg_obj)
 put_object_l(fn,o)
 puts(fn,end_obj)
end procedure


function get_object_l( integer fn )
 atom sz, a
 sequence r
 r = {}
 sz = getc(fn)
 while sz != end_obj do
  if sz = uchar then
   r &= getc(fn)
  elsif sz = char then
   a = getc(fn)
   r &= a-(256*(and_bits(a,#80)!=0))
  elsif sz = word then
   r &= get2u(fn)
  elsif sz = short then
   r &= get2s(fn)
  elsif sz = dword then
   r &= get4u(fn)
  elsif sz = long then
   r &= get4s(fn)
  elsif sz = double then
   r &= get_float64(fn)
  elsif sz = string then
   a = get4u(fn)
   r &= {getchar({fn,a})}
  elsif sz = beg_seq then
   r &= {get_object_l(fn)}
  elsif sz = end_seq then
   return r
  end if
  sz = getc(fn)
 end while
 return r
end function


global function get_object(integer fn)
 sequence s
 if getc(fn) = beg_obj then
  s = get_object_l(fn)
  return s[1]
 end if
 return -1
end function

-------------------------------------------------
-- v0.3a --
-- added:
--    file_size() -- does not move file-pointer
--    load_file() -- returns a pointer to file in memory
--
-------------------------------------------------
-- v0.2a --
-- added:
--    for recursive calls
--       put_object_l()
--       get_object_l()
-- modified:
--    found bug in the following routines, and modified them
--    to use recursive calls
--       put_object()
--       get_object()
-- fixed:
--       get2u()  -- was returning signed integers
-------------------------------------------------
-- v0.1b --
-- modified: (to prevent namespace conflicts)
--       seek()   -- renamed as seekfn()
--       where()  -- renamed as wherefn()
-------------------------------------------------
-- v0.1a --
--    removed all file routines from base.e and removed dependancies
--    on other include files
--       getchar()      -- modified version of getc()
--       put2()
--       put4()
--       put_float32()
--       put_float64()
--       getst()  -- modified gets
--       get2u()
--       get2s()
--       get4u()
--       get4s()
--       get_float32()
--       get_float64()
--       get_string()   -- modified version of gets()
--       byte_string()  -- local type
--       atom_size()    -- local (for put_object)
--       put_object()
--       get_object()

